#!/usr/bin/env python
# coding: utf-8


"""

======PROJECT======

MathEnergy

======DESCRIPTION======

The functions in this script are used to create absolute generation and load profiles of elements in a Simbench net object.

======Overview functions=====

01: def get_absolute_profiles_from_relative_profiles(net, element,
                                                     multiplying_column, 
                                                     relative_profiles = None, 
                                                     profile_column = "profile",
                                                     profile_suffix = None, 
                                                     time_as_index = False, 
                                                     **kwargs)

02: def get_absolute_values(net, **kwargs)


​
======AUTHOR, CONTACT & DATE======

Copyright (c) 2016-2020 by University of Kassel and Fraunhofer Institute for 
Energy Economics and Energy System Technology (IEE), Kassel. All rights reserved.
Modified by the Department of Energy and Resource Management, Technical 
University of Berlin.

verwiebe@tu-berlin.de

June 2021

"""



# ---IMPORT STATEMENTS---
import simbench as sb
import numpy as np
import pandas as pd

import logging
logger = logging.getLogger('mathenergy')



# ---FUNCTIONS---
# 01
def get_absolute_profiles_from_relative_profiles(net, element, 
                                                 multiplying_column, 
                                                 relative_profiles = None, 
                                                 profile_column = "profile",
                                                 profile_suffix = None, 
                                                 time_as_index = False, 
                                                 **kwargs):
    """Returns a dataframe with profiles for the given element (e.g. 'loads' or 
    'sgens'). The profile values are calculated by multiplying the relative 
    profiles given by relative_profiles (or if not given, from net["profiles"]) 
    with the values in net[element][multiplying_column].

    Parameters
    ----------
    net : (:obj:pandapowerNet)
        A pandapower grid.
    
    element : (str)
        Type for which absolute profiles are calculated. Possible are 'load', 
        'gen', 'sgen' or 'storage'.
    
    multiplying_column : (str)
        Column name within net[element].columns which should be multiplied with 
        the relative profiles. Usual multiply_columns are 'p_mw' or 'q_mvar'.
        Additional Feature: If multiplying_column is not a string, the relative 
        profiles are not multiplied with net[element][multiplying_column] but 
        with 'multiplying_column' itself.
    
    relative_profiles : (pd.DataFrame), optional
        Dataframe of relative profiles as input. If None, net["profiles"] is 
        considered. The default is None.
    
    profile_column : (str), optional
        Name of the column which contains information about which element is 
        assigned to which profile. In SimBench grids, this information is
        given in the column "profile". For that reason, 'profile' is the default.
    
    profile_suffix : (str), optional
        In case that different profiles are given for p and q, they can be 
        distinguished by a suffix. For loads this can be "_pload" and "_qload",
        which will be automatically assumed, if profile_suffix is None.
    
    time_as_index : (bool), optional
        If TRUE, the returned dataframe has relative_profiles["time"] as index. 
        If FALSE, relative_profiles.index is used. The default is FALSE.
    
    **kwargs
        Key word arguments for merge_dataframes().

    Raises
    ------
    ValueError

    Returns
    -------
    output_profiles : (pd.DataFrame)
        Calculated absolute profiles.

    """
    # Use net.profiles if relative_profiles is None
    if relative_profiles is None:
        if element in ["load", "storage", "dcline"]:
            relative_profiles = net.profiles[element]
        elif element in ["gen", "sgen"]:
            # Since RES and Powerplants can be converted to pandapower as both, gen or sgen, both are considered together
            relative_profiles = sb.merge_dataframes(
                [net.profiles["powerplants"], net.profiles["renewables"]], **kwargs)
        else:
            raise ValueError("element %s is unknown." % str(element))

    # Set index
    index = relative_profiles["time"] if time_as_index else relative_profiles.index
    if "time" in relative_profiles:
        del relative_profiles["time"]

    # Set profile_suffix assumptions if profile_suffix is None
    if profile_suffix is None:
        if element == "load":
            if multiplying_column == "p_mw":
                profile_suffix = "_pload"
            elif multiplying_column == "q_mvar":
                profile_suffix = "_qload"
        profile_suffix = "" if profile_suffix is None else profile_suffix

    # Get relative profiles with respect to each element index
    if profile_column in net[element].columns:
        applied_profiles = net[element][profile_column] + profile_suffix
    else:  # Missing profile column
        logger.warning("In %s table, profile column '%s' is missing. Scalings of 1 are assumed." % (
            element, profile_column))
        missing_col_handling = "missing_col_handling"
        applied_profiles = pd.Series([missing_col_handling] * net[element].shape[0],
                                     index = net[element].index)
        relative_profiles[missing_col_handling] = 1

    # NaN profile handling
    if applied_profiles.isnull().any():
        logger.debug("There are Nan profiles. Scalings of 1 are assumed.")
        nan_profile_handling = "nan_profile_scaling"
        assert nan_profile_handling not in relative_profiles.columns
        applied_profiles.loc[applied_profiles.isnull()] = nan_profile_handling
        relative_profiles[nan_profile_handling] = 1

    relative_output_profiles = relative_profiles.values[:, sb.idx_in_2nd_array(
	                                                       applied_profiles.values, 
                                                           np.array(relative_profiles.columns))]

    # Get factor to multiply with (consider additional feature of 'multiplying_column')
    if isinstance(multiplying_column, str):
        if multiplying_column in net[element].columns:
            factor = net[element][multiplying_column].values.reshape(1, -1)
        else:
            raise ValueError("'multiplying_column' %s is not net[%s].columns." % (
                multiplying_column, element))
    else:
        factor = multiplying_column

    # Multiply relative profiles with factor and return results
    output_profiles = pd.DataFrame(relative_output_profiles * factor, index = index,
                                   columns = net[element].index)
    return output_profiles
	
	
# 02
def get_absolute_values(net, **kwargs):
    """ Is a convenience function using 
    get_absolute_profiles_from_relative_profiles(). 
    This function returns a dictionary with all absolute values, calculated 
    from scaling factors and maximum active or reactive powers.

    Parameters
    ----------
    net : (:obj:pandapowerNet)
        A pandapower grid.
    
    **kwargs
        Key word arguments for get_absolute_profiles_from_relative_profiles()
        (especially for merge_dataframes()).

    Returns
    -------
    abs_val : (dict)
        Absolute values calculated from relative scaling factors and maximum
        active or reactive powers. The keys of this dictionary are tuples 
        consisting of element and parameter. The values are dataframes with 
        absolute power values.

    """
    abs_val = dict()

    # Use given profiles
    for elm_col in [("load", "p_mw"), ("load", "q_mvar"), ("sgen", "p_mw"), 
                    ("gen", "p_mw"), ("storage", "p_mw"), ("dcline","p_mw")]:
        abs_val[elm_col] = get_absolute_profiles_from_relative_profiles(
            net, elm_col[0], elm_col[1], **kwargs)

    return abs_val