#!/usr/bin/env python
# coding: utf-8


"""

======PROJECT======

MathEnergy

======DESCRIPTION======

Functions in this script are used to create an interactive geoplot of the Simbench net object.

======Overview functions=====

01: def _in_ipynb()

02: def sum_line_length(pts)

03: def get_line_neutral(coord)

04: def create_edge_center_trace(line_trace, size = 1, patch_type = "circle", 
                                 color = "white", infofunc = None,
                                 trace_name = 'edge_center', 
                                 use_line_geodata = False)

05: def create_bus_trace(net, buses = None, size = 5, patch_type = "circle", 
                         color = "blue", infofunc = None, trace_name = 'buses', 
                         legendgroup = None, cmap = None, cmap_vals = None, 
                         cbar_title = None, cmin = None, cmax = None, 
                         cpos = 1.0, colormap_column = "vm_pu")

06: def _get_line_geodata_plotly(net, lines, use_line_geodata, dclines = False)

07: def create_line_trace(net, lines = None, dclines = False, 
                          use_line_geodata = True, respect_switches = False, 
                          width = 1.0, color = 'grey', infofunc = None, 
                          trace_name = 'lines', legendgroup = None,
                          cmap = None, cbar_title = None, show_colorbar = True, 
                          cmap_vals = None, cmin = None, cmax = None, 
                          cpos = 1.1)

08: def draw_traces(traces, on_map = False, map_style = 'basic', 
                    showlegend = True, figsize = 1, aspectratio = 'auto', 
                    filename = "temp-plot.html")


​
======AUTHOR, CONTACT & DATE======

Copyright (c) 2016-2020 by University of Kassel and Fraunhofer Institute for 
Energy Economics and Energy System Technology (IEE), Kassel. All rights reserved.
Edited by the Department of Energy and Resource Management, Technical 
University of Berlin.

Tbd

June 2021

"""



# ---IMPORT STATEMENTS---
import math

import numpy as np
import pandas as pd
from packaging import version
from collections.abc import Iterable

from pandapower.plotting.plotly.get_colors import get_plotly_color, get_plotly_cmap
from pandapower.plotting.plotly.mapbox_plot import _on_map_test, _get_mapbox_token, \
    MapboxTokenMissing

try:
    import pplog as logging
except ImportError:
    import logging
logger = logging.getLogger(__name__)

try:
    from plotly import __version__ as plotly_version
    from plotly.graph_objs.scatter.marker import ColorBar
    from plotly.graph_objs import Figure, Layout
    from plotly.graph_objs.layout import XAxis, YAxis
    from plotly.graph_objs.scatter import Line, Marker
    from plotly.graph_objs.scattermapbox import Line as scmLine
    from plotly.graph_objs.scattermapbox import Marker as scmMarker
except ImportError:
    logger.info("Failed to import plotly - interactive plotting will not be available")



# ---FUNCTIONS---
# 01
def _in_ipynb():
    """An auxiliary function which checks if plot is called from a 
    jupyter-notebook or not.
    
    """
    import __main__ as main
    return not hasattr(main, '__file__')


# 02
def sum_line_length(pts):
    """Sums length of lines from start to end point.

    Parameters
    ----------
    pts : (list)
        List of x,y coordinates.

    Returns
    -------
    line_length : (list)
        List of line lengths.

    """
    pt_diff = lambda p: (p[0][0] - p[1][0], p[0][1] - p[1][1])
    diffs = map(pt_diff, zip(pts[:-1], pts[1:]))
    line_length = sum(math.hypot(d1, d2) for d1, d2 in diffs)
    return line_length


# 03
def get_line_neutral(coord):
    """Calculates middle point between a start and an end point

    Parameters
    ----------
    coord : (list)
        List of x,y coordinates.

    Returns
    -------
    mid: (list)
		List of middle points between coords-

    """
    if len(coord) == 1:
        return coord[0]
    half_length = sum_line_length(coord) / 2.0
    length = 0.0
    ind = 0
    while length < half_length:
        ind += 1
        length = sum_line_length(coord[:ind])

    start_coord = coord[ind - 2]
    end_coord = coord[ind - 1]
    mid = [(a1 + a2) / 2.0 for a1, a2 in zip(start_coord, end_coord)]

    return mid


# 04
def create_edge_center_trace(line_trace, size = 1, patch_type = "circle", 
                             color = "white", infofunc = None, 
                             trace_name = 'edge_center', 
                             use_line_geodata = False):
    """Creates a plotly trace of pandapower buses.

    Parameters
    ----------
    line_trace : (from pandapowerNet)
        The already generated line traces with center geodata.
    
    size : (int), optional
        Patch size. The default is 1.
    
    patch_type : (str), optional
        Patch type, can be:
            - "circle" for a circle,
            - "square" for a rectangle,
            - "diamond" for a diamond,
            - Much more pathc types at https://plot.ly/python/reference/#scatter-marker.
        The default is "circle".
    
    color : (str), optional
        Color of buses in the trace. The default is "white".
    
    infofunc : (pd.Series), optional
        Hoverinfo for each trace element. Indices should correspond to the 
        pandapower element indices. The default is None.
    
    trace_name : (str), optional
        Name of the trace which will appear in the legend. The default 
        is 'edge_center'.
    
    use_line_geodata : (bool), optional
        DESCRIPTION. The default is FALSE.

    Returns
    -------
    center_trace : (dict)
        Plotly trace of pandapower buses.

    """
    # color = get_plotly_color(color)

    center_trace = dict(type = 'scatter', text = [], mode = 'markers', hoverinfo = 'text', 
                        name = trace_name, marker = dict(color = color, size = size, 
                                                         symbol = patch_type))

    if not use_line_geodata:
        center_trace['x'], center_trace['y'] = (line_trace[0]["x"][1::4], 
                                                line_trace[0]["y"][1::4])
    else:
        x, y = [], []
        for trace in line_trace:
            coord = list(zip(trace["x"], trace["y"]))
            mid_coord = get_line_neutral(coord)
            x.append(mid_coord[0])
            y.append(mid_coord[1])

        center_trace['x'], center_trace['y'] = (x, y)

    center_trace['text'] = infofunc

    return center_trace


# 05
def create_bus_trace(net, buses = None, size = 5, patch_type = "circle", 
                     color = "blue", infofunc = None, trace_name = 'buses', 
                     legendgroup = None, cmap = None, cmap_vals = None, 
                     cbar_title = None, cmin = None, cmax = None, cpos = 1.0, 
                     colormap_column = "vm_pu"):
    """ Creates a plotly trace of pandapower buses.

    Parameters
    ----------
    net : (:obj:pandapowerNet)
        A pandapower grid.
    
    buses : (list), optional
        The buses for which the collections are created. If None, all buses in 
        the network are considered. The default is None.
    
    size : (int), optional
        Patch size. The default is 5.
    
    patch_type : (str), optional
        Patch type, can be:
            - "circle" for a circle,
            - "square" for a rectangle,
            - "diamond" for a diamond,
            - Much more pathc types at https://plot.ly/python/reference/#scatter-marker. 
        The default is "circle".
    
    color : (str), optional
        Color of buses in the trace. The default is "blue".
    
    infofunc : (pd.Series), optional
        Hoverinfo for bus elements. Indices should correspond to the pandapower 
        element indices. The default is None.
    
    trace_name : (str), optional
        Name of the trace which will appear in the legend. The default is 'buses'.
    
    legendgroup : (str), optional
        Defines groups of layers that will be displayed in a legend, 
        e.g. groups according to voltage level (as used in `vlevel_plotly`). 
        The default is None.
    
    cmap : (str), optional
        Name of a colormap which exists within plotly (Greys, YlGnBu, Greens, 
        YlOrRd, Bluered, RdBu, Reds, Blues, Picnic, Rainbow, Portland, Jet, Hot, 
        Blackbody, Earth, Electric, Viridis). Alternatively a custom discrete 
        colormap can be used. The default is None.
    
    cmap_vals : (list), optional
        Values used for coloring using colormap. The default is None.
    
    cbar_title : (str), optional
        Title for the colorbar. The default is None.
    
    cmin : (float), optional
        Colorbar range minimum. The default is None.
    
    cmax : (float), optional
        Colorbar range maximum. The default is None.
    
    cpos : (float), optional
        Position of the colorbar. The default is 1.0.
    
    colormap_column : (str), optional
        Set color of bus according to this variable. The default is "vm_pu".

    Returns
    -------
    list
        Plotly trace of pandapower buses.

    """
    color = get_plotly_color(color)

    bus_trace = dict(type='scatter', text = [], mode = 'markers', hoverinfo = 'text', 
                     name = trace_name, marker = dict(color = color, size = size, 
                                                      symbol = patch_type))

    buses = net.bus.index.tolist() if buses is None else list(buses)
    bus_plot_index = [b for b in buses if b in list(set(buses) & set(net.bus_geodata.index))]

    bus_trace['x'], bus_trace['y'] = (net.bus_geodata.loc[bus_plot_index, 'x'].tolist(),
                                      net.bus_geodata.loc[bus_plot_index, 'y'].tolist())

    if not isinstance(infofunc, pd.Series) and isinstance(infofunc, Iterable) and \
            len(infofunc) == len(buses):
        infofunc = pd.Series(index=buses, data = infofunc)

    bus_trace['text'] = net.bus.loc[bus_plot_index, 'name'] if infofunc is None else \
        infofunc.loc[buses]

    if legendgroup:
        bus_trace['legendgroup'] = legendgroup

    # If color map is set
    if cmap is not None:
        # If cmap_vals are not given

        cmap = 'Jet' if cmap is True else cmap

        if cmap_vals is not None:
            cmap_vals = cmap_vals
        else:
            if net.res_line.shape[0] == 0:
                logger.error("There are no power flow results for buses voltage magnitudes which are default for bus "
                             "colormap coloring..."
                             "set cmap_vals input argument if you want colormap according to some specific values...")
            cmap_vals = net.res_bus.loc[bus_plot_index, colormap_column].values

        cmap_vals = net.res_bus.loc[bus_plot_index, colormap_column] if cmap_vals is None else cmap_vals

        cmin = cmin if cmin else cmap_vals.min()
        cmax = cmax if cmax else cmap_vals.max()

        bus_trace['marker'] = Marker(size = size,
                                     color = cmap_vals, cmin = cmin, cmax = cmax,
                                     colorscale = cmap,
                                     colorbar = ColorBar(thickness = 10,
                                                         x = cpos),
                                     symbol = patch_type
                                     )

        if cbar_title:
            bus_trace['marker']['colorbar']['title'] = cbar_title

        bus_trace['marker']['colorbar']['title']['side'] = 'right'

    return [bus_trace]


# 06
def _get_line_geodata_plotly(net, lines, use_line_geodata, dclines = False):
    """Gets line geodata from line elements in Simbench network object (net).

    Parameters
    ----------
    net : (:obj:pandapowerNet)
        A pandapower grid.
    
    lines : (list), optional
        The lines for which the collections are created.
        If None, all lines in the network are considered. The default is None.
    
    use_line_geodata : (bool)
        If True, line geodata is used.
    
    dclines : (bool), optional
        Flag if the line trace is for AC lines (FALSE) or DC lines (TRUE). 
        The default is FALSE.

    Returns
    -------
    xs: (list)
        List of x coordinates of lines.
    
    ys: (list)
        List of y coordinates of lines.

    """
    xs = []
    ys = []
    if use_line_geodata & (not dclines):
        for line_ind, _ in lines.iterrows():
            line_coords = net.line_geodata.loc[line_ind, 'coords']
            linex, liney = list(zip(*line_coords))
            xs += linex
            xs += [None]
            ys += liney
            ys += [None]
    else:
        # Getting x and y values from bus_geodata for from and to side of each line

        from_bus = net.bus_geodata.loc[lines.from_bus, 'x'].tolist()
        to_bus = net.bus_geodata.loc[lines.to_bus, 'x'].tolist()
        # Center point added because of the hovertool
        center = (np.array(from_bus) + np.array(to_bus)) / 2
        none_list = [None] * len(from_bus)
        xs = np.array([from_bus, center, to_bus, none_list]).T.flatten().tolist()

        from_bus = net.bus_geodata.loc[lines.from_bus, 'y'].tolist()
        to_bus = net.bus_geodata.loc[lines.to_bus, 'y'].tolist()
        # Center point added because of the hovertool
        center = (np.array(from_bus) + np.array(to_bus)) / 2
        none_list = [None] * len(from_bus)
        ys = np.array([from_bus, center, to_bus, none_list]).T.flatten().tolist()

    # [:-1] is because the trace will not appear on maps if None is at the end
    return xs[:-1], ys[:-1]


# 07
def create_line_trace(net, lines = None, dclines = False, 
                      use_line_geodata = True, respect_switches = False, 
                      width = 1.0, color = 'grey', infofunc = None, 
                      trace_name = 'lines', legendgroup = None, cmap = None, 
                      cbar_title = None, show_colorbar = True, 
                      cmap_vals = None, cmin = None, cmax = None, cpos = 1.1):
    """Creates a plotly trace of pandapower lines.
    
    Parameters
    ----------
    net : (:obj:pandapowerNet)
        A pandapower grid.
    
    lines : (list), optional
        The lines for which the collections are created.
        If None, all lines in the network are considered. The default is None.
    
    dclines : (bool), optional
        Flag if the line trace is for AC lines (FALSE) or DC lines (TRUE). 
        The default is FALSE.
    
    use_line_geodata : TYPE, optional
        DESCRIPTION. The default is TRUE.
    
    respect_switches : (bool), optional
        Flag for consideration of disconnected lines. The default is FALSE.
    
    width : (int), optional
        Line width. The default is 1.0.
    
    color : (str), optional
        Color of lines in the trace. The default is 'grey'.
    
    infofunc : (pd.Series), optional
        Hoverinfo for line elements. Indices should correspond to the 
        pandapower element indices. The default is None.
    
    trace_name : (str), optional
        Name of the trace which will appear in the legend. 
        The default is 'lines'.
    
    legendgroup : (str), optional
        Defines groups of layers that will be displayed in a legend, 
        e.g. groups according to voltage level (as used in `vlevel_plotly`). 
        The default is None.
    
    cmap : (str), optional
        Name of a colormap which exists within plotly if set to TRUE default 
        'Jet' colormap is used. Alternative colormaps : Greys, YlGnBu, Greens, 
        YlOrRd, Bluered, RdBu, Reds, Blues, Picnic, Rainbow, Portland, Jet, 
        Hot, Blackbody, Earth, Electric, Viridis. The default is None.
    
    cbar_title : (str), optional
        Title for the colorbar. The default is None.
    
    show_colorbar : (bool), optional
        Flag for showing or not corresponding colorbar. The default is TRUE.
    
    cmap_vals : (list), optional
        Values used for coloring using colormap. The default is None.
    
    cmin : (float), optional
        Colorbar range minimum. The default is None.
    
    cmax : (float), optional
        Colorbar range maximum. The default is None.
    
    cpos : (float), optional
        Position of the colorbar. The default is 1.1.

    Raises
    ------
    UserWarning

    Returns
    -------
    line_traces : (dict)
        Plotly trace of pandapower lines.

    """
    color = get_plotly_color(color)

    # Defining lines to be plot
    lines = net.line.index.tolist() if lines is None else list(lines)
    if len(lines) == 0:
        return []

    if infofunc is not None:
        if not isinstance(infofunc, pd.Series) and isinstance(infofunc, Iterable) and \
                len(infofunc) == len(lines):
            infofunc = pd.Series(index = lines, data = infofunc)
        if len(infofunc) != len(lines) and len(infofunc) != len(net.line):
            raise UserWarning("Different amount of hover info than lines to plot")
        assert isinstance(infofunc, pd.Series), \
            "infofunc should be a pandas series with the net.line.index to the infofunc contents"

    no_go_lines = set()
    if respect_switches:
        no_go_lines = set(lines) & set(net.switch.element[(net.switch.et == "l") &
                                                          (net.switch.closed == 0)])
        if dclines & len(no_go_lines):
            print("Be careful if this lines are switched off (no_go_lines): ", no_go_lines)

    if dclines:
        lines_to_plot = net.dcline.loc[set(net.dcline.index) & set(lines)]
    else:
        lines_to_plot = net.line.loc[set(net.line.index) & (set(lines) - no_go_lines)]
    no_go_lines_to_plot = None
    if not dclines:
        use_line_geodata = use_line_geodata if net.line_geodata.shape[0] > 0 else False
    else:
        use_line_geodata = False
        

    if use_line_geodata:
        lines_to_plot = lines_to_plot.loc[set(lines_to_plot.index) & set(net.line_geodata.index)]
    else:
        lines_with_geodata = lines_to_plot.from_bus.isin(net.bus_geodata.index) & \
                             lines_to_plot.to_bus.isin(net.bus_geodata.index)
        lines_to_plot = lines_to_plot.loc[lines_with_geodata]

    cmap_lines = None
    if cmap is not None:
        # Workaround: If colormap plot is used, each line needs to be separate scatter object because
        # plotly still doesn't support appropriately colormap for line objects
        # TODO --> Correct this when plotly solves existing github issue about Line colorbar

        cmap = 'jet' if cmap is True else cmap

        if cmap_vals is not None:
            if not isinstance(cmap_vals, np.ndarray):
                cmap_vals = np.asarray(cmap_vals)
        else:
            if net.res_line.shape[0] == 0:
                logger.error("There are no power flow results for lines which are default for line colormap coloring..."
                             "set cmap_vals input argument if you want colormap according to some specific values...")
            if dclines:
                cmap_vals = net.res_dcline.loc[lines_to_plot.index, 'loading_percent'].values
            else:
                cmap_vals = net.res_line.loc[lines_to_plot.index, 'loading_percent'].values

        cmap_lines = get_plotly_cmap(cmap_vals, cmap_name = cmap, cmin = cmin, cmax = cmax)
        if len(cmap_lines) == len(net.line):
            # Some lines are not plotted although cmap_value were provided for all lines
            line_idx_map = dict(zip(net.line.loc[lines].index.tolist(), range(len(lines))))
            cmap_lines = [cmap_lines[line_idx_map[idx]] for idx in lines_to_plot.index]
        else:
            assert len(cmap_lines) == len(lines_to_plot), \
                "Different amounts of cmap values and lines to plot were supplied"

    line_traces = []
    for col_i, (idx, line) in enumerate(lines_to_plot.iterrows()):
        line_color = color
        line_info = line['name']
        if cmap is not None:
            try:
                line_color = cmap_lines[col_i]
                line_info = line['name'] if infofunc is None else infofunc.loc[idx]
            except IndexError:
                logger.warning("No color and info for line {:d} (name: {}) available".format(
                    idx, line['name']))

        line_trace = dict(type = 'scatter', text = [], hoverinfo = 'text', mode = 'lines', 
                          name = trace_name, line = Line(width = width, color = color))

        line_trace['x'], line_trace['y'] = _get_line_geodata_plotly(net, 
                                                                    lines_to_plot.loc[idx:idx],
                                                                    use_line_geodata, dclines)

        line_trace['line']['color'] = line_color

        line_trace['text'] = line_info

        line_traces.append(line_trace)

    if show_colorbar and cmap is not None:

        cmin = cmin if cmin else cmap_vals.min()
        cmax = cmax if cmax else cmap_vals.max()
        try:
            # TODO --> For custom colormaps
            cbar_cmap_name = 'Jet' if cmap == 'jet' else cmap
            # Workaround to get colorbar for lines (an unvisible node is added)
            # Get x and y of first line.from_bus:
            if dclines:
                x = [net.bus_geodata.x[net.dcline.from_bus[net.dcline.index[0]]]]
                y = [net.bus_geodata.y[net.dcline.from_bus[net.dcline.index[0]]]]
            else: 
                x = [net.bus_geodata.x[net.line.from_bus[net.line.index[0]]]]
                y = [net.bus_geodata.y[net.line.from_bus[net.line.index[0]]]]
            lines_cbar = dict(type = 'scatter', x = x, y = y, mode = 'markers',
                              marker=Marker(size = 0, cmin = cmin, cmax = cmax,
                                            color = 'rgb(255,255,255)',
                                            opacity = 0,
                                            colorscale = cbar_cmap_name,
                                            colorbar = ColorBar(thickness = 10,
                                                                x = cpos),
                                            ))
            if cbar_title:
                lines_cbar['marker']['colorbar']['title'] = cbar_title

            lines_cbar['marker']['colorbar']['title']['side'] = 'right'

            line_traces.append(lines_cbar)
        except:
            pass

    if (len(no_go_lines) > 0) & (not dclines):
        no_go_lines_to_plot = net.line.loc[no_go_lines]
        for idx, line in no_go_lines_to_plot.iterrows():
            line_color = color
            line_trace = dict(type = 'scatter', text = [], hoverinfo = 'text', 
                              mode = 'lines', name = 'disconnected lines',
                              line = Line(width = width / 2, color = 'grey', 
                                          dash = 'dot'))

            line_trace['x'], line_trace['y'] = _get_line_geodata_plotly(
                net, no_go_lines_to_plot.loc[idx:idx], use_line_geodata)

            line_trace['line']['color'] = line_color
            try:
                line_trace['text'] = infofunc.loc[idx]
            except (KeyError, IndexError, AttributeError):
                line_trace["text"] = line['name']

            line_traces.append(line_trace)

            if legendgroup:
                line_trace['legendgroup'] = legendgroup

    # Sort infofunc so that it is the correct order lines_to_plot + no_go_lines_to_plot
    if infofunc is not None:
        if dclines: 
            if not isinstance(infofunc, pd.Series) and isinstance(infofunc, Iterable) and \
                len(infofunc) == len(net.dcline):
                infofunc = pd.Series(index = net.dcline.index, data = infofunc)
        else:
            if not isinstance(infofunc, pd.Series) and isinstance(infofunc, Iterable) and \
                len(infofunc) == len(net.line):
                infofunc = pd.Series(index = net.line.index, data = infofunc)
        assert isinstance(infofunc, pd.Series), \
            "infofunc should be a pandas series with the net.line.index to the infofunc contents"
        sorted_idx = lines_to_plot.index.tolist()
        if no_go_lines_to_plot is not None:
            sorted_idx += no_go_lines_to_plot.index.tolist()
        infofunc = infofunc.loc[sorted_idx]

    center_trace = create_edge_center_trace(line_traces, color = color, 
                                            infofunc = infofunc,
                                            use_line_geodata = use_line_geodata)
    line_traces.append(center_trace)
    
    return line_traces


# 08
def draw_traces(traces, on_map = False, map_style = 'basic', showlegend = True, 
                figsize = 1, aspectratio = 'auto', filename = "temp-plot.html"):
    """Plots all the traces (which can be created using 
    :func:'create_bus_trace' or :func:'create_line_trace')
    to PLOTLY (see https://plot.ly/python/).

    Parameters
    ----------
    traces : (list of dicts)
        List of dicts which correspond to plotly traces generated using: 
        'create_bus_trace' or 'create_line_trace'.
    
    on_map : (bool), optional
        Enables using mapbox plot in plotly. The default is FALSE.
    
    map_style : (str), optional
        Enables using mapbox plot in plotly:
            - 'streets'
            - 'bright'
            - 'light'
            - 'dark'
            - 'satellite'. 
        The default is 'basic'.
    
    showlegend : (bool), optional
        Enables legend display. The default is TRUE.
    
    figsize : (float), optional
        Aspectratio is multiplied by it in order to get final image size. 
        The default is 1.
    
    aspectratio : (tuple), optional
        When 'auto' it preserves original aspect ratio of them network geodata 
        any custom aspectration can be given as a tuple, e.g. (1.2, 1). 
        The default is 'auto'.
    
    filename : (str), optional
        Plots to a html file called filename. The default is "temp-plot.html".

    Raises
    ------
    MapboxTokenMissing

    Returns
    -------
    fig : (graph_objs._figure.Figure)
        Figure object.

    """
    if on_map:
        try:
            on_map = _on_map_test(traces[0]['x'][0], traces[0]['y'][0])
        except:
            logger.warning("Test if geo-data are in lat/long cannot be performed using geopy -> "
                           "eventual plot errors are possible.")

        if on_map is False:
            logger.warning("Existing geodata are not real lat/lon geographical coordinates. -> "
                           "plot on maps is not possible.\n"
                           "Use geo_data_to_latlong(net, projection) to transform geodata from specific projection.")

    if on_map:
        # Change traces for mapbox
        # Change trace_type to scattermapbox and rename x to lat and y to lon
        for trace in traces:
            trace['lat'] = trace.pop('x')
            trace['lon'] = trace.pop('y')
            trace['type'] = 'scattermapbox'
            if "line" in trace and isinstance(trace["line"], Line):
                # Scattermapboxplot lines do not support dash for some reason, make it a red line instead
                if "dash" in trace["line"]._props:
                    _prps = dict(trace["line"]._props)
                    _prps.pop("dash", None)
                    _prps["color"] = "red"
                    trace["line"] = scmLine(_prps)
                else:
                    trace["line"] = scmLine(dict(trace["line"]._props))
            elif "marker" in trace and isinstance(trace["marker"], Marker):
                trace["marker"] = scmMarker(trace["marker"]._props)

    # Setting Figure object
    fig = Figure(data = traces,  # Edge_trace
                 layout = Layout(
                     titlefont = dict(size = 16),
                     showlegend = showlegend,
                     autosize = True if aspectratio == 'auto' else False,
                     hovermode = 'closest',
                     margin = dict(b=5, l=5, r=5, t=5),
                     # Annotations=[dict(
                     #     text="",
                     #     showarrow=False,
                     #     xref="paper", yref="paper",
                     #     x=0.005, y=-0.002)],
                     xaxis = XAxis(showgrid = False, zeroline = False, showticklabels = False),
                     yaxis = YAxis(showgrid = False, zeroline = False, showticklabels = False),
                     # legend=dict(x=0, y=1.0)
                 ), )

    # Check if geodata are real geographycal lat/lon coordinates using geopy

    if on_map:
        try:
            mapbox_access_token = _get_mapbox_token()
        except Exception:
            logger.exception('mapbox token required for map plots. '
                             'Get Mapbox token by signing in to https://www.mapbox.com/.\n'
                             'After getting a token, set it to pandapower using:\n'
                             'pandapower.plotting.plotly.mapbox_plot.set_mapbox_token(\'<token>\')')
            raise MapboxTokenMissing

        fig['layout']['mapbox'] = dict(accesstoken = mapbox_access_token,
                                       bearing = 0,
                                       center = dict(lat = pd.Series(traces[0]['lat']).dropna().mean(),
                                                     lon = pd.Series(traces[0]['lon']).dropna().mean()),
                                       style = map_style,
                                       pitch = 0,
                                       zoom = 11)

    # Default aspectratio: If on_map use auto, else use 'original'
    aspectratio = 'original' if not on_map and aspectratio == 'auto' else aspectratio

    if aspectratio != 'auto':
        if aspectratio == 'original':
            # TODO --> Improve this workaround for getting original aspectratio
            xs = []
            ys = []
            for trace in traces:
                xs += trace['x']
                ys += trace['y']
            x_dropna = pd.Series(xs).dropna()
            y_dropna = pd.Series(ys).dropna()
            xrange = x_dropna.max() - x_dropna.min()
            yrange = y_dropna.max() - y_dropna.min()
            ratio = xrange / yrange
            if ratio < 1:
                aspectratio = (ratio, 1.)
            else:
                aspectratio = (1., 1 / ratio)

        aspectratio = np.array(aspectratio) / max(aspectratio)
        fig['layout']['width'], fig['layout']['height'] = ([ar * figsize * 700 for ar in aspectratio])

    # Check if called from ipynb or not in order to consider appropriate plot function
    if _in_ipynb():
        from plotly.offline import init_notebook_mode, iplot as plot
        init_notebook_mode()
    else:
        from plotly.offline import plot as plot

    plot(fig, filename = filename)

    return fig